"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRiskStats = exports.getEquityForMangoAccounts = exports.getPerpPositionsToBeLiquidated = exports.getPriceImpactForLiqor = exports.getOnChainPriceForMints = exports.computePriceImpactOnJup = void 0;
const web3_js_1 = require("@solana/web3.js");
const bn_js_1 = __importDefault(require("bn.js"));
const cloneDeep_1 = __importDefault(require("lodash/cloneDeep"));
const mangoAccount_1 = require("./accounts/mangoAccount");
const I80F48_1 = require("./numbers/I80F48");
const utils_1 = require("./utils");
async function buildFetch() {
    let fetch = globalThis?.fetch;
    if (!fetch && process?.versions?.node) {
        fetch = (await Promise.resolve().then(() => __importStar(require('node-fetch')))).default;
    }
    return fetch;
}
async function computePriceImpactOnJup(amount, inputMint, outputMint) {
    const url = `https://quote-api.jup.ag/v4/quote?inputMint=${inputMint}&outputMint=${outputMint}&amount=${amount}&swapMode=ExactIn&slippageBps=10000&onlyDirectRoutes=false&asLegacyTransaction=false`;
    const response = await (await buildFetch())(url, { mode: 'no-cors' });
    try {
        const res = await response.json();
        if (res['data'] && res.data.length > 0 && res.data[0].outAmount) {
            return {
                outAmount: parseFloat(res.data[0].outAmount),
                priceImpactPct: parseFloat(res.data[0].priceImpactPct),
            };
        }
        else {
            return {
                outAmount: -1 / 10000,
                priceImpactPct: -1 / 10000,
            };
        }
    }
    catch (e) {
        console.log(e);
        return {
            outAmount: -1 / 10000,
            priceImpactPct: -1 / 10000,
        };
    }
}
exports.computePriceImpactOnJup = computePriceImpactOnJup;
async function getOnChainPriceForMints(mints) {
    return await Promise.all(mints.map(async (mint) => {
        const resp = await (await buildFetch())(`https://public-api.birdeye.so/public/price?address=${mint}`, {
            headers: {
                'Content-Type': 'application/json',
            },
        });
        const data = await resp.json();
        return data?.data?.value;
    }));
}
exports.getOnChainPriceForMints = getOnChainPriceForMints;
async function getPriceImpactForLiqor(group, mangoAccounts) {
    const mangoAccountsWithHealth = mangoAccounts.map((a) => {
        return {
            account: a,
            health: a.getHealth(group, mangoAccount_1.HealthType.liquidationEnd),
            healthRatio: a.getHealthRatioUi(group, mangoAccount_1.HealthType.liquidationEnd),
        };
    });
    const usdcBank = group.getFirstBankByTokenIndex(0);
    const usdcMint = usdcBank.mint;
    return await Promise.all(Array.from(group.banksMapByMint.values())
        .sort((a, b) => a[0].name.localeCompare(b[0].name))
        .map(async (banks) => {
        const bank = banks[0];
        // Sum of all liabs, these liabs would be acquired by liqor,
        // who would immediately want to reduce them to 0
        // Assuming liabs need to be bought using USDC
        const liabs = 
        // Max liab of a particular token that would be liquidated to bring health above 0
        mangoAccountsWithHealth.reduce((sum, a) => {
            // How much would health increase for every unit liab moved to liqor
            // liabprice * (liabweight - (1+fee)*assetweight)
            // Choose the most valuable asset the user has
            const assetBank = Array.from(group.banksMapByTokenIndex.values())
                .flat()
                .reduce((prev, curr) => prev.initAssetWeight
                .mul(a.account.getEffectiveTokenBalance(group, prev))
                .mul(prev._price)
                .gt(curr.initAssetWeight.mul(a.account
                .getEffectiveTokenBalance(group, curr)
                .mul(curr._price)))
                ? prev
                : curr);
            const tokenLiabHealthContrib = bank.price.mul(bank.initLiabWeight.sub((0, I80F48_1.ONE_I80F48)()
                .add(bank.liquidationFee)
                .mul(assetBank.initAssetWeight)));
            // Abs liab/borrow
            const maxTokenLiab = a.account
                .getEffectiveTokenBalance(group, bank)
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs();
            if (tokenLiabHealthContrib.eq((0, I80F48_1.ZERO_I80F48)())) {
                return sum.add(maxTokenLiab);
            }
            // Health under 0
            const maxLiab = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(tokenLiabHealthContrib)
                .min(maxTokenLiab);
            return sum.add(maxLiab);
        }, (0, I80F48_1.ZERO_I80F48)());
        const liabsInUsdc = 
        // convert to usdc, this is an approximation
        liabs
            .mul(bank.price)
            .floor()
            // jup oddity
            .min(I80F48_1.I80F48.fromNumber(99999999999));
        // Sum of all assets which would be acquired in exchange for also acquiring
        // liabs by the liqor, who would immediately want to reduce to 0
        // Assuming assets need to be sold to USDC
        const assets = mangoAccountsWithHealth.reduce((sum, a) => {
            // How much would health increase for every unit liab moved to liqor
            // assetprice * (liabweight/(1+liabliqfee) - assetweight)
            // Choose the smallest liability the user has
            const liabBank = Array.from(group.banksMapByTokenIndex.values())
                .flat()
                .reduce((prev, curr) => prev.initLiabWeight
                .mul(a.account.getEffectiveTokenBalance(group, prev))
                .mul(prev._price)
                .lt(curr.initLiabWeight.mul(a.account
                .getEffectiveTokenBalance(group, curr)
                .mul(curr._price)))
                ? prev
                : curr);
            const tokenAssetHealthContrib = bank.price.mul(liabBank.initLiabWeight
                .div((0, I80F48_1.ONE_I80F48)().add(liabBank.liquidationFee))
                .sub(bank.initAssetWeight));
            // Abs collateral/asset
            const maxTokenHealthAsset = a.account
                .getEffectiveTokenBalance(group, bank)
                .max((0, I80F48_1.ZERO_I80F48)());
            if (tokenAssetHealthContrib.eq((0, I80F48_1.ZERO_I80F48)())) {
                return sum.add(maxTokenHealthAsset);
            }
            const maxAsset = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(tokenAssetHealthContrib)
                .min(maxTokenHealthAsset);
            return sum.add(maxAsset);
        }, (0, I80F48_1.ZERO_I80F48)());
        const [pi1, pi2] = await Promise.all([
            !liabsInUsdc.eq((0, I80F48_1.ZERO_I80F48)()) &&
                usdcMint.toBase58() !== bank.mint.toBase58()
                ? computePriceImpactOnJup(liabsInUsdc.toString(), usdcMint.toBase58(), bank.mint.toBase58())
                : Promise.resolve({ priceImpactPct: 0, outAmount: 0 }),
            !assets.eq((0, I80F48_1.ZERO_I80F48)()) &&
                usdcMint.toBase58() !== bank.mint.toBase58()
                ? computePriceImpactOnJup(assets.floor().toString(), bank.mint.toBase58(), usdcMint.toBase58())
                : Promise.resolve({ priceImpactPct: 0, outAmount: 0 }),
        ]);
        return {
            Coin: { val: bank.name, highlight: false },
            'Oracle Price': {
                val: bank['oldUiPrice'] ? bank['oldUiPrice'] : bank._uiPrice,
                highlight: false,
            },
            'Jup Price': {
                val: bank['onChainPrice'],
                highlight: Math.abs((bank['onChainPrice'] -
                    (bank['oldUiPrice'] ? bank['oldUiPrice'] : bank._uiPrice)) /
                    (bank['oldUiPrice'] ? bank['oldUiPrice'] : bank._uiPrice)) > 0.05,
            },
            'Future Price': { val: bank._uiPrice, highlight: false },
            'V4 Liq Fee': {
                val: Math.round(bank.liquidationFee.toNumber() * 10000),
                highlight: false,
            },
            Liabs: {
                val: Math.round((0, utils_1.toUiDecimalsForQuote)(liabsInUsdc)),
                highlight: Math.round((0, utils_1.toUiDecimalsForQuote)(liabsInUsdc)) > 5000,
            },
            'Liabs Slippage': {
                val: Math.round(pi1.priceImpactPct * 10000),
                highlight: Math.round(pi1.priceImpactPct * 10000) >
                    Math.round(bank.liquidationFee.toNumber() * 10000),
            },
            Assets: {
                val: Math.round((0, utils_1.toUiDecimals)(assets, bank.mintDecimals) * bank.uiPrice),
                highlight: Math.round((0, utils_1.toUiDecimals)(assets, bank.mintDecimals) * bank.uiPrice) > 5000,
            },
            'Assets Slippage': {
                val: Math.round(pi2.priceImpactPct * 10000),
                highlight: Math.round(pi2.priceImpactPct * 10000) >
                    Math.round(bank.liquidationFee.toNumber() * 10000),
            },
        };
    }));
}
exports.getPriceImpactForLiqor = getPriceImpactForLiqor;
async function getPerpPositionsToBeLiquidated(group, mangoAccounts) {
    const mangoAccountsWithHealth = mangoAccounts.map((a) => {
        return {
            account: a,
            health: a.getHealth(group, mangoAccount_1.HealthType.liquidationEnd),
            healthRatio: a.getHealthRatioUi(group, mangoAccount_1.HealthType.liquidationEnd),
        };
    });
    return Array.from(group.perpMarketsMapByMarketIndex.values())
        .filter((pm) => !pm.name.includes('OLD'))
        .map((pm) => {
        const baseLots = mangoAccountsWithHealth
            .filter((a) => a.account.getPerpPosition(pm.perpMarketIndex))
            .reduce((sum, a) => {
            const baseLots = a.account.getPerpPosition(pm.perpMarketIndex).basePositionLots;
            const unweightedHealthPerLot = baseLots.gt(new bn_js_1.default(0))
                ? I80F48_1.I80F48.fromNumber(-1)
                    .mul(pm.price)
                    .mul(I80F48_1.I80F48.fromU64(pm.baseLotSize))
                    .mul(pm.initBaseAssetWeight)
                    .add(I80F48_1.I80F48.fromU64(pm.baseLotSize)
                    .mul(pm.price)
                    .mul((0, I80F48_1.ONE_I80F48)() // quoteInitAssetWeight
                    .mul((0, I80F48_1.ONE_I80F48)().sub(pm.baseLiquidationFee))))
                : pm.price
                    .mul(I80F48_1.I80F48.fromU64(pm.baseLotSize))
                    .mul(pm.initBaseLiabWeight)
                    .sub(I80F48_1.I80F48.fromU64(pm.baseLotSize)
                    .mul(pm.price)
                    .mul((0, I80F48_1.ONE_I80F48)()) // quoteInitLiabWeight
                    .mul((0, I80F48_1.ONE_I80F48)().add(pm.baseLiquidationFee)));
            const maxBaseLots = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(unweightedHealthPerLot.abs())
                .min(I80F48_1.I80F48.fromU64(baseLots).abs());
            return sum.add(maxBaseLots);
        }, (0, I80F48_1.ONE_I80F48)());
        const notionalPositionUi = (0, utils_1.toUiDecimalsForQuote)(baseLots.mul(I80F48_1.I80F48.fromU64(pm.baseLotSize).mul(pm.price)));
        return {
            Market: { val: pm.name, highlight: false },
            Price: { val: pm['oldUiPrice'], highlight: false },
            'Future Price': { val: pm._uiPrice, highlight: false },
            'Notional Position': {
                val: Math.round(notionalPositionUi),
                highlight: Math.round(notionalPositionUi) > 5000,
            },
        };
    });
}
exports.getPerpPositionsToBeLiquidated = getPerpPositionsToBeLiquidated;
async function getEquityForMangoAccounts(client, group, mangoAccounts) {
    // Filter mango accounts which might be closed
    const liqors = (await client.connection.getMultipleAccountsInfo(mangoAccounts))
        .map((ai, i) => {
        return { ai: ai, pk: mangoAccounts[i] };
    })
        .filter((val) => val.ai)
        .map((val) => val.pk);
    const liqorMangoAccounts = await Promise.all(liqors.map((liqor) => client.getMangoAccount(liqor, true)));
    const accountsWithEquity = liqorMangoAccounts.map((a) => {
        return {
            Account: { val: a.publicKey, highlight: false },
            Equity: {
                val: Math.round((0, utils_1.toUiDecimalsForQuote)(a.getEquity(group))),
                highlight: false,
            },
        };
    });
    accountsWithEquity.sort((a, b) => b.Equity.val - a.Equity.val);
    return accountsWithEquity;
}
exports.getEquityForMangoAccounts = getEquityForMangoAccounts;
async function getRiskStats(client, group, change = 0.4) {
    // Get known liqors
    let liqors;
    try {
        liqors = (await (await (await buildFetch())(`https://api.mngo.cloud/data/v4/stats/liqors-over_period?over_period=1MONTH`, { mode: 'no-cors' })).json()).map((data) => new web3_js_1.PublicKey(data['liqor']));
    }
    catch (error) {
        liqors = [];
    }
    // Get known mms
    const mms = [
        new web3_js_1.PublicKey('BLgb4NFwhpurMrGX5LQfb8D8dBpGSGtBqqew2Em8uyRT'),
        new web3_js_1.PublicKey('DA5G4mUdFmQXyKuFqvGGZGBzPAPYDYQsdjmiEJAYzUXQ'),
        new web3_js_1.PublicKey('BGYWnqfaauCeebFQXEfYuDCktiVG8pqpprrsD4qfqL53'),
        new web3_js_1.PublicKey('F1SZxEDxxCSLVjEBbMEjDYqajWRJQRCZBwPQnmcVvTLV'),
    ];
    // Get all mango accounts
    const mangoAccounts = await client.getAllMangoAccounts(group, true);
    // const mangoAccounts = [
    //   await client.getMangoAccount(
    //     new PublicKey('5G9XriaoqQy1V4s9RmnbczWAozzbv6h2RuEeAHk4R6Lb'), // https://app.mango.markets/stats?token=SOL
    //     true,
    //   ),
    // ];
    // Get on chain prices
    const mints = [
        ...new Set(Array.from(group.banksMapByTokenIndex.values())
            .flat()
            .map((bank) => bank.mint.toString())),
    ];
    const prices = await getOnChainPriceForMints([
        ...new Set(Array.from(group.banksMapByTokenIndex.values())
            .flat()
            .map((bank) => bank.mint.toString())),
    ]);
    const onChainPrices = Object.fromEntries(prices.map((price, i) => [mints[i], price]));
    Array.from(group.banksMapByTokenIndex.values())
        .flat()
        .forEach((b) => {
        b['onChainPrice'] = onChainPrices[b.mint.toBase58()];
    });
    // Clone group, and simulate change % price drop for all assets except stables
    const drop = 1 - change;
    const groupDrop = (0, cloneDeep_1.default)(group);
    Array.from(groupDrop.banksMapByTokenIndex.values())
        .flat()
        .filter((b) => !b.name.includes('USD'))
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * drop;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    Array.from(groupDrop.perpMarketsMapByMarketIndex.values()).forEach((p) => {
        p['oldUiPrice'] = p._uiPrice;
        p._uiPrice = p._uiPrice * drop;
        p._price = p._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    // Clone group, and simulate change % price drop for usdc
    const groupUsdcDepeg = (0, cloneDeep_1.default)(group);
    Array.from(groupUsdcDepeg.banksMapByTokenIndex.values())
        .flat()
        .filter((b) => b.name.includes('USDC'))
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * drop;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    // Clone group, and simulate change % price drop for usdt
    const groupUsdtDepeg = (0, cloneDeep_1.default)(group);
    Array.from(groupUsdtDepeg.banksMapByTokenIndex.values())
        .flat()
        .filter((b) => b.name.includes('USDT'))
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * drop;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    // Clone group, and simulate change % price rally for all assets except stables
    const rally = 1 + change;
    const groupRally = (0, cloneDeep_1.default)(group);
    Array.from(groupRally.banksMapByTokenIndex.values())
        .flat()
        .filter((b) => !b.name.includes('USD'))
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * rally;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(rally));
    });
    Array.from(groupRally.perpMarketsMapByMarketIndex.values()).forEach((p) => {
        p['oldUiPrice'] = p._uiPrice;
        p._uiPrice = p._uiPrice * rally;
        p._price = p._price?.mul(I80F48_1.I80F48.fromNumber(rally));
    });
    const [assetDrop, assetRally, usdcDepeg, usdtDepeg, perpDrop, perpRally, liqorEquity, marketMakerEquity,] = await Promise.all([
        getPriceImpactForLiqor(groupDrop, mangoAccounts),
        getPriceImpactForLiqor(groupRally, mangoAccounts),
        getPriceImpactForLiqor(groupUsdcDepeg, mangoAccounts),
        getPriceImpactForLiqor(groupUsdtDepeg, mangoAccounts),
        getPerpPositionsToBeLiquidated(groupDrop, mangoAccounts),
        getPerpPositionsToBeLiquidated(groupRally, mangoAccounts),
        getEquityForMangoAccounts(client, group, liqors),
        getEquityForMangoAccounts(client, group, mms),
    ]);
    return {
        assetDrop: {
            title: `Table 1a: Liqors acquire liabs and assets. The assets and liabs are sum of max assets and max
    liabs for any token which would be liquidated to fix the health of a mango account.
    This would be the slippage they would face on buying-liabs/offloading-assets tokens acquired from unhealth accounts after a 40% drop to all non-stable oracles`,
            data: assetDrop,
        },
        assetRally: {
            title: `Table 1b: ... same as above but with a 40% rally to all non-stable oracles instead of drop`,
            data: assetRally,
        },
        usdcDepeg: {
            title: `Table 1c: ... same as above but with a 40% drop to only usdc oracle`,
            data: usdcDepeg,
        },
        usdtDepeg: {
            title: `Table 1d: ... same as above but with a 40% drop to only usdt oracle`,
            data: usdtDepeg,
        },
        perpDrop: {
            title: `Table 2a: Perp notional that liqor need to liquidate after a  40% drop`,
            data: perpDrop,
        },
        perpRally: {
            title: `Table 2b: Perp notional that liqor need to liquidate after a  40% rally`,
            data: perpRally,
        },
        liqorEquity: {
            title: `Table 3: Equity of known liqors from last month`,
            data: liqorEquity,
        },
        marketMakerEquity: {
            title: `Table 4: Equity of known makers from last month`,
            data: marketMakerEquity,
        },
    };
}
exports.getRiskStats = getRiskStats;
